open Ast_typer

type infos = {
  init_addr : int;
  methods : (string, int) Hashtbl.t;
  constructors : int list;
  class_addr : int;
}

let infos_of_class_name : (string, infos) Hashtbl.t = Hashtbl.create 100
let addr_of_class_name = Hashtbl.create 100
let tc_of_class_name = Hashtbl.create 100
  
let search_label label list =
  let p (lab, _, _) = lab = label in
    List.find p list
;;

let compute_var_ind block_sizes expr_ofs bid vid =
  let rec f bsl bid acc =
    match (bsl, bid) with
      | ([], _) -> assert false
      | (bs :: tl, 0) -> acc + bs - vid + expr_ofs - 1
      | (bs :: tl, _) -> f tl (pred bid) (bs + acc)
  in
    f block_sizes bid 0
;;

let compute_nb_field cname =
  let rec f tc acc =
    let n = List.length tc.tc_class.typed_class_fields in
      match tc.tc_father with
	| None -> (n + acc)
	| Some tcf -> f tcf (n + acc)
  in
    f (Hashtbl.find tc_of_class_name cname) 0
;;

let compute_field_ind typed_class cid fid =
  let rec compute_tc tc depth =
    if depth = 0 then tc else
      match tc.tc_father with
	| None -> assert false
	| Some tcf -> compute_tc tcf (pred depth)
  in
  let rec compute_ofs tco acc =
    match tco with
      | None -> acc
      | Some tc ->
	  let field_nb = List.length tc.tc_class.typed_class_fields in
	    compute_ofs tc.tc_father (acc + field_nb)
  in
    compute_ofs (compute_tc typed_class cid).tc_father 0 + fid + 1;
;;

let get_method_addrs tc super_id =
  let rec f tco n =
    match tco with 
      | None -> assert false
      | Some tc ->
	  if n = 1 then
	    (Hashtbl.find infos_of_class_name tc.tc_name).methods
	  else
	    f tc.tc_father (pred n)
  in
    f tc.tc_father super_id
;;

let rec iter_tc f tc =
  f tc;
  List.iter (iter_tc f) tc.tc_children;
;;

let store_class_decls typed_class =
  let store_tc tc = Hashtbl.add tc_of_class_name tc.tc_name tc in
    iter_tc store_tc typed_class
;;

let get_nb_method cname =
  let tc = Hashtbl.find tc_of_class_name cname in
    tc.tc_method_count
;;

let get_addr_of_class_name cname =
  try Hashtbl.find addr_of_class_name cname
  with Not_found ->
    let addr = ref (-1) in
      Hashtbl.add addr_of_class_name cname addr;
      addr
;;

let fill_class_addrs () =
  let f cname fake_addr =
    fake_addr := (Hashtbl.find infos_of_class_name cname).class_addr;
  in
    Hashtbl.iter f addr_of_class_name;
;;
